@@ -1,3 +1,4 @@
+
 =head1 NAME
 
 PGObject - A toolkit integrating intelligent PostgreSQL dbs into Perl objects
@@ -7,20 +8,20 @@ PGObject - A toolkit integrating intelligent PostgreSQL dbs into Perl objects
 package PGObject;
 use strict;
 use warnings;
+
 use Carp;
 use Memoize;
+use PGObject::Type::Registry;
+
+use List::MoreUtils qw/pairwise/;
 
 =head1 VERSION
 
-Version 1.403.2
+Version 2.2.0
 
 =cut
 
-our $VERSION = '1.403.2';
-
-my %typeregistry = (
-    default => {},
-);
+our $VERSION = '2.2.0';
 
 =head1 SYNPOSIS
 
@@ -30,7 +31,7 @@ To use without caching:
 
 To use with caching:
 
-  use PGObject ':cache'; 
+  use PGObject ':cache';
 
 To get basic info from a function
 
@@ -76,6 +77,8 @@ To do the same with a running total
 sub import {
     my @directives = @_;
     memoize 'function_info' if grep { $_ eq ':cache' } @directives;
+    PGObject::Type::Registry->new_registry($_)
+        for grep { $_ !~ /^\:/; } @directives;
 }
 
 =head1 DESCRIPTION
@@ -85,14 +88,14 @@ stored procedures in PostgreSQL databases.  This module contains only common
 functionality and support structures, and low-level API's.  Most developers will
 want to use more functional modules which add to these functions.
 
-The overall approach here is to provide the basics for a toolkit that other 
-modules can extend.  This is thus intended to be a component for building 
-integration between PostgreSQL user defined functions and Perl objects.  
+The overall approach here is to provide the basics for a toolkit that other
+modules can extend.  This is thus intended to be a component for building
+integration between PostgreSQL user defined functions and Perl objects.
 
-Because decisions such as state handling are largely outside of the scope of 
-this module, this module itself does not do any significant state handling.  
-Database handles (using DBD::Pg 2.0 or later) must be passed in on every call. 
-This decision was made in order to allow for diversity in this area, with the 
+Because decisions such as state handling are largely outside of the scope of
+this module, this module itself does not do any significant state handling.
+Database handles (using DBD::Pg 2.0 or later) must be passed in on every call.
+This decision was made in order to allow for diversity in this area, with the
 idea that wrapper classes would be written to implement this.
 
 =head1 FUNCTIONS
@@ -111,7 +114,6 @@ sub clear_info_cache {
     eval { Memoize::flush_cache('function_info') };
 }
 
-
 =head2 function_info(%args)
 
 Arguments:
@@ -128,7 +130,7 @@ function name
 
 =item funcschema (optional, default 'public')
 
-function schema 
+function schema
 
 =item funcprefix (optiona, default '')
 
@@ -144,8 +146,8 @@ Name of first argument type's schema.  If not provided defaults to 'public'
 
 =back
 
-This function looks up basic mapping information for a function.  If more than 
-one function is found, an exception is raised.  This function is primarily 
+This function looks up basic mapping information for a function.  If more than
+one function is found, an exception is raised.  This function is primarily
 intended to be used by packages which extend this one, in order to accomplish
 stored procedure to object mapping.
 
@@ -157,7 +159,7 @@ Return data is a hashref containing the following elements:
 
 This is an arrayref of hashrefs, each of which contains 'name' and 'type'
 
-=item name 
+=item name
 
 The name of the function
 
@@ -170,57 +172,54 @@ The number of arguments
 =cut
 
 sub function_info {
-    my ($self) = shift @_;
-    my %args = @_;
+    my ( $self, %args ) = @_;
     $args{funcschema} ||= 'public';
     $args{funcprefix} ||= '';
-    $args{funcname} = $args{funcprefix}.$args{funcname};
+    $args{funcname} = $args{funcprefix} . $args{funcname};
     $args{argschema} ||= 'public';
 
-    my $dbh = $args{dbh};
-
-    
+    my $dbh = $args{dbh} || croak 'No dbh provided';
 
     my $query = qq|
-        SELECT proname, pronargs, proargnames, 
-               string_to_array(array_to_string(proargtypes::regtype[], ' '), 
+        SELECT proname, pronargs, proargnames,
+               string_to_array(array_to_string(proargtypes::regtype[], ' '),
                                ' ') as argtypes
-          FROM pg_proc 
+          FROM pg_proc
           JOIN pg_namespace pgn ON pgn.oid = pronamespace
          WHERE proname = ? AND nspname = ?
     |;
-    my @queryargs = ($args{funcname}, $args{funcschema});
-    if ($args{argtype1}) {
-       $query .= qq|
-               AND (proargtypes::int[])[0] IN (select t.oid 
+    my @queryargs = ( $args{funcname}, $args{funcschema} );
+    if ( $args{argtype1} ) {
+        $query .= qq|
+               AND (proargtypes::int[])[0] IN (select t.oid
                                                  from pg_type t
                                                  join pg_namespace n
                                                       ON n.oid = typnamespace
-                                                where typname = ? 
+                                                where typname = ?
                                                       AND n.nspname = ?
        )|;
-       push @queryargs, $args{argtype1};
-       push @queryargs, $args{argschema};
+        push @queryargs, $args{argtype1};
+        push @queryargs, $args{argschema};
     }
 
     my $sth = $dbh->prepare($query) || die $!;
     $sth->execute(@queryargs) || die $dbh->errstr . ": " . $query;
     my $ref = $sth->fetchrow_hashref('NAME_lc');
-    croak "transaction already aborted" if  $dbh->state eq '25P02';
-    croak "No such function" if !$ref;
+    croak "transaction already aborted"  if $dbh->state eq '25P02';
+    croak "No such function"             if !$ref;
     croak 'Ambiguous discovery criteria' if $sth->fetchrow_hashref('NAME_lc');
 
     my $f_args;
-    for my $n (@{$ref->{proargnames}}){
-        push @$f_args, {name => $n, type => shift @{$ref->{argtypes}}};
+    for my $n ( @{ $ref->{proargnames} } ) {
+        push @$f_args, { name => $n, type => shift @{ $ref->{argtypes} } };
     }
 
     return {
-        name     => $ref->{proname}, 
+        name     => $ref->{proname},
         num_args => $ref->{pronargs},
         args     => $f_args,
     };
-    
+
 }
 
 =head2 call_procedure(%args)
@@ -243,8 +242,8 @@ Prefix for the function.  This can be useful for separating functions by class.
 
 =item args
 
-This is an arrayref.  Each item is either a literal value, an arrayref, or a 
-hashref of extended information.  In the hashref case, the type key specifies 
+This is an arrayref.  Each item is either a literal value, an arrayref, or a
+hashref of extended information.  In the hashref case, the type key specifies
 the string to use to cast the type in, and value is the value.
 
 =item orderby
@@ -255,14 +254,14 @@ The list (arrayref) of columns on output for ordering.
 
 An arrayref of running windowed aggregates.  Each contains two keys, namely 'agg' for the aggregate and 'alias' for the function name.
 
-These are aggregates, each one has appended 'OVER (ROWS UNBOUNDED PRECEDING)' 
-to it.  
+These are aggregates, each one has appended 'OVER (ROWS UNBOUNDED PRECEDING)'
+to it.
 
 =item registry
 
 This is the name of the registry used for type conversion.  It can be omitted
-and defaults to 'default.'  Note that use of a non-standard registry currently 
-does *not* merge changes from the default registry, so you need to reregister 
+and defaults to 'default.'  Note that use of a non-standard registry currently
+does *not* merge changes from the default registry, so you need to reregister
 types in non-default registries when you create them.
 
 Please note, these aggregates are not intended to be user-supplied.  Please only
@@ -275,58 +274,61 @@ the framework level for this parameter.
 =cut
 
 sub call_procedure {
-    my ($self) = shift @_;
-    my %args = @_;
+    my ( $self, %args ) = @_;
     local $@;
     $args{funcschema} ||= 'public';
     $args{funcprefix} ||= '';
-    $args{funcname} = $args{funcprefix}.$args{funcname};
+    $args{funcname} = $args{funcprefix} . $args{funcname};
     $args{registry} ||= 'default';
 
-    my $registry = $typeregistry{$args{registry}};
     my $dbh = $args{dbh};
     croak "No database handle provided" unless $dbh;
-    croak "dbh not a database handle" unless eval {$dbh->isa('DBI::db')};
+    croak "dbh not a database handle" unless eval { $dbh->isa('DBI::db') };
 
     my $wf_string = '';
 
-    $wf_string = join ', ', map { 
-                         $_->{agg} 
-                         . ' OVER (ROWS UNBOUNDED PRECEDING) AS '
-                         . $_->{alias}
-                 } @{$args{running_funcs}} if $args{running_funcs};
+    $wf_string = join ', ', map {
+              $_->{agg}
+            . ' OVER (ROWS UNBOUNDED PRECEDING) AS '
+            . $_->{alias}
+    } @{ $args{running_funcs} } if $args{running_funcs};
     $wf_string = ', ' . $wf_string if $wf_string;
 
-    my @qargs = map { 
-                      my $arg = $_;
-                      local ($@);
-                      $arg = $arg->to_db if eval {$arg->can('to_db')};
-                      $arg = $arg->pgobject_to_db if eval {$arg->can('pgobject_to_db')};
-                      $arg;
-                }  @{$args{args}};
+    my @qargs = map {
+        my $arg = $_;
+        local ($@);
+        $arg = $arg->to_db          if eval { $arg->can('to_db') };
+        $arg = $arg->pgobject_to_db if eval { $arg->can('pgobject_to_db') };
+        $arg;
+    } @{ $args{args} };
 
-    my $argstr = join ', ', map { 
-                  (ref $_ and eval { $_->{cast} } ) ? "?::$_->{cast}" : '?';
-                  } @{$args{args}};  
+    my $argstr = join ', ', map {
+        ( ref $_ and eval { $_->{cast} } ) ? "?::$_->{cast}" : '?';
+    } @{ $args{args} };
 
     my $order = '';
-    if ($args{orderby}){
-        $order = join(', ', map {
-                                  my $dir = undef;
-                                  if ( s/\s+(ASC|DESC)\s*$//i ) {
-                                      $dir = $1;
-                                  }
-                                  defined $dir ? $dbh->quote_identifier($_)
-                                                  . " $dir"
-                                               : $dbh->quote_identifier($_);
-                                } @{$args{orderby}}); 
+    if ( $args{orderby} ) {
+        $order = join(
+            ', ',
+            map {
+                my $dir = undef;
+                if (s/\s+(ASC|DESC)\s*$//i) {
+                    $dir = $1;
+                }
+                defined $dir
+                    ? $dbh->quote_identifier($_) . " $dir"
+                    : $dbh->quote_identifier($_);
+            } @{ $args{orderby} }
+        );
     }
     my $query = qq|
-           SELECT * $wf_string 
-             FROM | . $dbh->quote_identifier($args{funcschema}) . '.' . 
-                      $dbh->quote_identifier($args{funcname}) . qq|($argstr) |;
-    if ($order){ 
-       $query .= qq|
+           SELECT * $wf_string
+             FROM |
+        . $dbh->quote_identifier( $args{funcschema} ) . '.'
+        . $dbh->quote_identifier( $args{funcname} )
+        . qq|($argstr) |;
+    if ($order) {
+        $query .= qq|
          ORDER BY $order |;
     }
 
@@ -334,79 +336,44 @@ sub call_procedure {
 
     my $place = 1;
 
-    foreach my $carg (@qargs){
-        if (ref($carg) =~ /HASH/){
-            $sth->bind_param($place, $carg->{value},
-                       { pg_type => $carg->{type} });
-        } else {
+    foreach my $carg (@qargs) {
+        if ( ref($carg) =~ /HASH/ ) {
+            $sth->bind_param( $place, $carg->{value},
+                { pg_type => $carg->{type} } );
+        }
+        else {
 
-            # This is used to support arrays of db-aware types.  Long-run 
+            # This is used to support arrays of db-aware types.  Long-run
             # I think we should merge bytea support into this framework. --CT
-            if (ref($carg) =~ /ARRAY/){
-               local ($@);
-               if (eval{$carg->[0]->can('to_db')}){
-                  for my $ref(@$carg){
-                       $ref = $ref->to_db;
-                  }
-               }
+            if ( ref($carg) =~ /ARRAY/ ) {
+                local ($@);
+                if ( eval { $carg->[0]->can('to_db') } ) {
+                    for my $ref (@$carg) {
+                        $ref = $ref->to_db;
+                    }
+                }
             }
 
-            $sth->bind_param($place, $carg);
+            $sth->bind_param( $place, $carg );
         }
         ++$place;
     }
 
     $sth->execute() || die $dbh->errstr . ": " . $query;
-    
-    clear_info_cache() if $dbh->state eq '42883'; # (No Such Function)
+
+    clear_info_cache() if $dbh->state eq '42883';    # (No Such Function)
 
     my @rows = ();
-    while (my $row = $sth->fetchrow_hashref('NAME_lc')){
-       my @types = @{$sth->{pg_type}};
-       my @names = @{$sth->{NAME_lc}};
-       my $i = 0;
-       for my $type (@types){
-           $row->{$names[$i]} 
-                 = process_type($row->{$names[$i]}, $type, $registry);
-           ++$i;
-       }
-       
-       push @rows, $row;
+    my $row_deserializer =
+        PGObject::Type::Registry->rowhash_deserializer(
+            registry => $args{registry},
+            types    => $sth->{pg_type},
+            columns  => $sth->{NAME_lc},
+        );
+    while (my $row = $sth->fetchrow_hashref('NAME_lc')) {
+        push @rows, $row_deserializer->( $row );
     }
-    return @rows;      
-}
-
-=head2 process_type($val, $type, $registry)
-
-If $type is registered, returns "$type"->from_db($val).  Otherwise returns
-$val.  If $val is an arrayref, loops through it for every item and strips 
-trialing [] from $type.
-
-This module should generally only be used by type handlers or by this module.
-
-=cut
-
-sub process_type {
-    my ($val, $type, $registry, $dbh) = @_;
-    return unless defined $val;
-
-    $registry = $typeregistry{$registry} unless ref $registry;
-    # Array handling as we'd get this usually from DBD::Pg or equivalent
-    if (ref $val and ref($val) =~ /ARRAY/){
-       # strangely, DBD::Pg returns, as of 2.x, the types of array types 
-       # as prefixed with an underscore.  So we have to remove this. --CT
-       $type =~ s/^\_//;
-       my $newval = [];
-       push @$newval, process_type($_, $type, $registry) for @$val;
-       return $newval;
-    }
-
-    # Otherwise:
-    if (defined $registry->{$type}){
-       my $class = $registry->{$type};
-       $val = $class->from_db($val);
-    }
-    return $val;
+    return @rows;
 }
 
 =head2 new_registry($registry_name)
@@ -414,19 +381,24 @@ sub process_type {
 Creates a new registry if it does not exist.  This is useful when segments of
 an application must override existing type mappings.
 
-Returns 1 on creation, 2 if already exists.
+This is deprecated and throws a warning.
+
+Use PGObject::Type::Registry->new_registry($registry_name) instead.
+
+This no longer returns anything of significance.
 
 =cut
 
-sub new_registry{
-    my ($self, $registry_name) = @_;
-    return 2 if defined $typeregistry{$registry_name};
-    $typeregistry{$registry_name} = {};
-    return 1;
+sub new_registry {
+    my ( $self, $registry_name ) = @_;
+    carp "Deprecated use of PGObject->new_registry()";
+    PGObject::Type::Registry->new_registry($registry_name);
 }
 
 =head2 register_type(pgtype => $tname, registry => $regname, perl_class => $pm)
 
+DEPRECATED
+
 Registers a type as a class.  This means that when an attribute of type $pg_type
 is returned, that PGObject will automatically return whatever
 $perl_class->from_db returns.  This allows you to have a db-specific constructor
@@ -437,41 +409,26 @@ The registry argument is optional and defaults to 'default'
 If the registry does not exist, an error is raised.  if the pg_type is already
 registered to a different type, this returns 0.  Returns 1 on success.
 
+Use PGObject::Type::Registry->register_type() instead.
+
 =cut
 
-sub register_type{
-    my $self = shift @_;
-    my %args = @_;
-    $args{registry} ||= 'default';
-    croak "Registry $args{registry} does not exist yet!" 
-              if !defined $typeregistry{$args{registry}};
-    return 0 if defined $typeregistry{$args{registry}}->{$args{pg_type}}
-             and $args{perl_class} 
-             ne $typeregistry{$args{registry}}->{$args{pg_type}};
-            
-    $typeregistry{$args{registry}}->{$args{pg_type}} = $args{perl_class};
+sub register_type {
+    carp 'Use of deprecated method register_type of PGObject module';
+    my ( $self, %args ) = @_;
+
+    PGObject::Type::Registry->register_type(
+        registry => $args{registry},
+        dbtype   => $args{pg_type},
+        apptype  => $args{perl_class}
+    );
     return 1;
 }
 
-=head2 get_registered(registry => $registry, pg_type => $pg_type)
-
-This is a public interface to the registry, which can be useful for composite
-types decoding themselves from tuple data, and so forth.
-
-=cut
-
-sub get_registered {
-    my ($self) = shift @_;
-    my %args = @_;
-    $args{registry} ||= 'default';
-    croak "Registry $args{registry} does not exist yet!"
-              if !defined $typeregistry{$args{registry}};
-    return undef unless defined $typeregistry{$args{registry}}->{$args{pg_type}};
-    return $typeregistry{$args{registry}}->{$args{pg_type}};
-}
-
 =head2 unregister_type(pgtype => $tname, registry => $regname)
 
+Deprecated.
+
 Tries to unregister the type.  If the type does not exist, returns 0, otherwise
 returns 1.  This is mostly useful for when a specific type must make sure it has
 the slot.  This is rarely desirable.  It is usually better to use a subregistry
@@ -479,33 +436,23 @@ instead.
 
 =cut
 
-sub unregister_type{
-    my $self = shift @_;
-    my %args = @_;
+sub unregister_type {
+    carp 'Use of deprecated method unregister_type of PGObject';
+    my ( $self, %args ) = @_;
+
     $args{registry} ||= 'default';
-    croak "Registry $args{registry} does not exist yet!" 
-              if !defined $typeregistry{$args{registry}};
-    return 0 if not defined $typeregistry{$args{registry}}->{$args{pg_type}};
-    delete $typeregistry{$args{registry}}->{$args{pg_type}};
-    return 1;
-}
-
-=head2 $hashref = get_type_registry()
-
-Returns the type registry.  Mostly useful for debugging.
-
-=cut
-
-sub get_type_registry {
-    return \%typeregistry;
+    PGObject::Type::Registry->unregister_type(
+        registry => $args{registry},
+        dbtype   => $args{pg_type}
+    );
 }
 
 =head1 WRITING PGOBJECT-AWARE HELPER CLASSES
 
 One of the powerful features of PGObject is the ability to declare methods in
 types which can be dynamically detected and used to serialize data for query
-purposes. Objects which contain a pgobject_to_db() or a to_db() method, that 
-method will be called and the return value used in place of the object.  This 
+purposes. Objects which contain a pgobject_to_db() or a to_db() method, that
+method will be called and the return value used in place of the object.  This
 can allow arbitrary types to serialize themselves in arbitrary ways.
 
 For example a date object could be set up with such a method which would export
@@ -523,16 +470,20 @@ returned, it must follow the type format:
 
 =head2 REQUIRED INTERFACES
 
-Registered types MUST implement a $class->from_db function accepts the string 
-from the database as its only argument, and returns the object of the desired 
+Registered types MUST implement a $class->from_db function accepts the string
+from the database as its only argument, and returns the object of the desired
 type.
 
-Any type MAY present an $object->to_db() interface, requiring no arguments, and returning a valid value.  These can be hashrefs as specified above, arrayrefs 
+Any type MAY present an $object->to_db() interface, requiring no arguments, and returning a valid value.  These can be hashrefs as specified above, arrayrefs
 (converted to PostgreSQL arrays by DBD::Pg) or scalar text values.
 
 =head2 UNDERSTANDING THE REGISTRY SYSTEM
 
-The registry system allows Perl classes to "claim" PostgreSQL types within a 
+Note that 2.0 moves the registry to a service module which handles both
+registry and deserialization of database types.  This is intended to be both
+cleaner and more flexible than the embedded system in 1.x.
+
+The registry system allows Perl classes to "claim" PostgreSQL types within a
 certain domain.  For example, if I want to ensure that all numeric types are
 turned into Math::BigFloat objects, I can build a wrapper class with appropriate
 interfaces, but PGObject won't know to convert numeric types to this new class,
@@ -549,6 +500,21 @@ specify non-standard registries when calling procedures, and PGObject will use
 only those components registered on the non-standard registry when checking rows
 before output.
 
+=head3 Backwards Incompatibilities from 1.x
+
+Deserialization occurs in a context which specifies a registry.  In 1.x there
+were no concerns about default mappings but now this triggers a warning.  The
+most basic and frequently used portions of this have been kept but return values
+for registering types has changed.  We no longer provide a return variable but
+throw an exception if the type cannot be safely registered.
+
+This follows a philosophy of throwing exceptions when guarantees cannot be met.
+
+We now throw warnings when the default registry is used.
+
+Longer-run, deserializers should use the PGObject::Type::Registry interface
+directly.
+
 =head1 WRITING TOP-HALF OBJECT FRAMEWORKS FOR PGOBJECT
 
 PGObject is intended to be the database-facing side of a framework for objects.
@@ -584,8 +550,8 @@ module and similar functions in other catalog-bound modules may not be safe to
 modify in arbitrary ways.  Therefore we recommend that the return values from
 catalog-lookup functions are treated as immutable.
 
-Normalizing output is safe provided there are no conflicts between naming 
-conventions.  This is usually true since different naming conventions would 
+Normalizing output is safe provided there are no conflicts between naming
+conventions.  This is usually true since different naming conventions would
 interfere withmapping.  However, there could be cases where it is not true, for
 example, where two different mapping modules agree on a subset of normalization
 conventions but differ on some details.  The two might safely handle the same
@@ -633,7 +599,7 @@ You can find documentation for this module with the perldoc command.
 
 You can also look for information at:
 
-=over 
+=over
 
 =item * RT: CPAN's request tracker (report bugs here)
 
@@ -655,12 +621,12 @@ L<http://search.cpan.org/dist/PGObject/>
 
 =head1 ACKNOWLEDGEMENTS
 
-This code has been loosely based on code written for the LedgerSMB open source 
+This code has been loosely based on code written for the LedgerSMB open source
 accounting and ERP project.  While that software uses the GNU GPL v2 or later,
-this is my own reimplementation, based on my original contributions to that 
+this is my own reimplementation, based on my original contributions to that
 project alone, and it differs in significant ways.   This being said, without
-LedgerSMB, this module wouldn't exist, and without the lessons learned there, 
-and the great people who have helped make this possible, this framework would 
+LedgerSMB, this module wouldn't exist, and without the lessons learned there,
+and the great people who have helped make this possible, this framework would
 not be half of what it is today.
 
 
@@ -677,27 +643,30 @@ not be half of what it is today.
 =head1 COPYRIGHT
 
 COPYRIGHT (C) 2013-2014 Chris Travers
+COPYRIGHT (C) 2014-2017 The LedgerSMB Core Team
 
-Redistribution and use in source and compiled forms with or without 
+Redistribution and use in source and compiled forms with or without
 modification, are permitted provided that the following conditions are met:
 
 =over
 
-=item 
+=item
 
 Redistributions of source code must retain the above
 copyright notice, this list of conditions and the following disclaimer as the
 first lines of this file unmodified.
 
-=item 
+=item
 
 Redistributions in compiled form must reproduce the above copyright
 notice, this list of conditions and the following disclaimer in the
-source code, documentation, and/or other materials provided with the 
+source code, documentation, and/or other materials provided with the
 distribution.
 
 =back
 
+=head1 LICENSE
+
 THIS SOFTWARE IS PROVIDED BY THE AUTHOR(S) "AS IS" AND
 ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
